@@ -91,6 +91,17 @@ def _validate_test(cls):
         raise ValueError(f'decorated test ({cls.__qualname__!r}) has one or '
                          f'more undefined parameters')
 
+    conditions = [VersionValidator(v) for v in cls._rfm_required_version]
+    if (cls._rfm_required_version and
+        not any(c.validate(osext.reframe_version()) for c in conditions)):
+
+        getlogger().warning(f"skipping incompatible test "
+                            f"'{cls.__qualname__}': not valid for ReFrame "
+                            f"version {osext.reframe_version().split('-')[0]}")
+        return False
+
+    return True
+
 
 def simple_test(cls):
     '''Class decorator for registering tests with ReFrame.
@@ -101,10 +112,9 @@ def simple_test(cls):
 
     .. versionadded:: 2.13
     '''
-    _validate_test(cls)
-
-    for _ in cls.param_space:
-        _register_test(cls)
+    if _validate_test(cls):
+        for _ in cls.param_space:
+            _register_test(cls)
 
     return cls
 
@@ -139,14 +149,14 @@ def parameterized_test(*inst):
     )
 
     def _do_register(cls):
-        _validate_test(cls)
-        if not cls.param_space.is_empty():
-            raise ValueError(
-                f'{cls.__qualname__!r} is already a parameterized test'
-            )
+        if _validate_test(cls):
+            if not cls.param_space.is_empty():
+                raise ValueError(
+                    f'{cls.__qualname__!r} is already a parameterized test'
+                )
 
-        for args in inst:
-            _register_test(cls, args)
+            for args in inst:
+                _register_test(cls, args)
 
         return cls
 
@@ -187,6 +197,12 @@ def required_version(*versions):
        These should be written as ``3.5.0`` and ``3.5.0-dev.0``.
 
     '''
+    warn.user_deprecation_warning(
+        "the '@required_version' decorator is deprecated; please set "
+        "the 'require_version' parameter in the class definition instead",
+        from_version='3.7.0'
+    )
+
     if not versions:
         raise ValueError('no versions specified')
 
@@ -198,8 +214,10 @@ def required_version(*versions):
             mod.__rfm_skip_tests = set()
 
         if not any(c.validate(osext.reframe_version()) for c in conditions):
-            getlogger().info('skipping incompatible test defined'
-                             ' in class: %s' % cls.__name__)
+            getlogger().warning(
+                f"skipping incompatible test '{cls.__qualname__}': not valid "
+                f"for ReFrame version {osext.reframe_version().split('-')[0]}"
+            )
             mod.__rfm_skip_tests.add(cls)
 
         return cls
