@@ -16,11 +16,7 @@ import reframe.core.variables as variables
 import reframe.core.hooks as hooks
 
 from reframe.core.exceptions import ReframeSyntaxError
-
-
-_USER_PIPELINE_STAGES = (
-    'init', 'setup', 'compile', 'run', 'sanity', 'performance', 'cleanup'
-)
+from reframe.core.deferrable import deferrable
 
 
 class RegressionTestMeta(type):
@@ -28,11 +24,11 @@ class RegressionTestMeta(type):
     class MetaNamespace(namespaces.LocalNamespace):
         '''Custom namespace to control the cls attribute assignment.
 
-        Regular Python class attributes can be overriden by either
+        Regular Python class attributes can be overridden by either
         parameters or variables respecting the order of execution.
         A variable or a parameter may not be declared more than once in the
         same class body. Overriding a variable with a parameter or the other
-        way around has an undefined behaviour. A variable's value may be
+        way around has an undefined behavior. A variable's value may be
         updated multiple times within the same class body. A parameter's
         value may not be updated more than once within the same class body.
         '''
@@ -175,6 +171,7 @@ class RegressionTestMeta(type):
         namespace['variable'] = variables.TestVar
         namespace['required'] = variables.Undefined
 
+        # Utility decorators
         def bind(fn, name=None):
             '''Directive to bind a free function to a class.
 
@@ -185,50 +182,48 @@ class RegressionTestMeta(type):
             namespace[inst.__name__] = inst
             return inst
 
+        def final(fn):
+            '''Indicate that a function is final and cannot be overridden.'''
+
+            fn._rfm_final = True
+            return fn
+
         namespace['bind'] = bind
+        namespace['final'] = final
 
         # Hook-related functionality
         def run_before(stage):
-            '''Decorator for attaching a test method to a pipeline stage.
+            '''Decorator for attaching a test method to a given stage.
 
             See online docs for more information.
             '''
-
-            if stage not in _USER_PIPELINE_STAGES:
-                raise ValueError(
-                    f'invalid pipeline stage specified: {stage!r}'
-                )
-
-            if stage == 'init':
-                raise ValueError('pre-init hooks are not allowed')
-
             return hooks.attach_to('pre_' + stage)
 
-        namespace['run_before'] = run_before
-
         def run_after(stage):
-            '''Decorator for attaching a test method to a pipeline stage.
+            '''Decorator for attaching a test method to a given stage.
 
             See online docs for more information.
             '''
-
-            if stage not in _USER_PIPELINE_STAGES:
-                raise ValueError(
-                    f'invalid pipeline stage specified: {stage!r}'
-                )
-
-            # Map user stage names to the actual pipeline functions if needed
-            if stage == 'init':
-                stage = '__init__'
-            elif stage == 'compile':
-                stage = 'compile_wait'
-            elif stage == 'run':
-                stage = 'run_wait'
-
             return hooks.attach_to('post_' + stage)
 
+        namespace['run_before'] = run_before
         namespace['run_after'] = run_after
         namespace['require_deps'] = hooks.require_deps
+
+        # Machinery to add a sanity function
+        def sanity_function(fn):
+            '''Mark a function as the test's sanity function.
+
+            Decorated functions must be unary and they will be converted into
+            deferred expressions.
+            '''
+
+            _def_fn = deferrable(fn)
+            setattr(_def_fn, '_rfm_sanity_fn', True)
+            return _def_fn
+
+        namespace['sanity_function'] = sanity_function
+        namespace['deferrable'] = deferrable
         return metacls.MetaNamespace(namespace)
 
     def __new__(metacls, name, bases, namespace, **kwargs):
@@ -241,11 +236,12 @@ class RegressionTestMeta(type):
         constructed.
         '''
 
-        blacklist = [
+        directives = [
             'parameter', 'variable', 'bind', 'run_before', 'run_after',
-            'require_deps', 'required'
+            'require_deps', 'required', 'deferrable', 'sanity_function',
+            'final'
         ]
-        for b in blacklist:
+        for b in directives:
             namespace.pop(b, None)
 
         return super().__new__(metacls, name, bases, dict(namespace), **kwargs)
@@ -255,9 +251,8 @@ class RegressionTestMeta(type):
 
         # Create a set with the attribute names already in use.
         cls._rfm_dir = set()
-        for base in bases:
-            if hasattr(base, '_rfm_dir'):
-                cls._rfm_dir.update(base._rfm_dir)
+        for base in (b for b in bases if hasattr(b, '_rfm_dir')):
+            cls._rfm_dir.update(base._rfm_dir)
 
         used_attribute_names = set(cls._rfm_dir)
 
@@ -275,26 +270,50 @@ class RegressionTestMeta(type):
         # attribute; all dependencies will be resolved first in the post-setup
         # phase if not assigned elsewhere
         hook_reg = hooks.HookRegistry.create(namespace)
-        for b in bases:
-            if hasattr(b, '_rfm_pipeline_hooks'):
-                hook_reg.update(getattr(b, '_rfm_pipeline_hooks'))
+        for base in (b for b in bases if hasattr(b, '_rfm_hook_registry')):
+            hook_reg.update(getattr(base, '_rfm_hook_registry'),
+                            denied_hooks=namespace)
+
+        cls._rfm_hook_registry = hook_reg
+
+        # Gather all the locally defined sanity functions based on the
+        # _rfm_sanity_fn attribute.
+
+        sn_fn = [v for v in namespace.values() if hasattr(v, '_rfm_sanity_fn')]
+        if sn_fn:
+            cls._rfm_sanity = sn_fn[0]
+            if len(sn_fn) > 1:
+                raise ReframeSyntaxError(
+                    f'{cls.__qualname__!r} defines more than one sanity '
+                    'function in the class body.'
+                )
+
+        else:
+            # Search the bases if no local sanity functions exist.
+            for base in (b for b in bases if hasattr(b, '_rfm_sanity')):
+                cls._rfm_sanity = getattr(base, '_rfm_sanity')
+                if cls._rfm_sanity.__name__ in namespace:
+                    raise ReframeSyntaxError(
+                        f'{cls.__qualname__!r} overrides the candidate '
+                        f'sanity function '
+                        f'{cls._rfm_sanity.__qualname__!r} without '
+                        f'defining an alternative'
+                    )
+
+                break
 
-        cls._rfm_pipeline_hooks = hook_reg
         cls._final_methods = {v.__name__ for v in namespace.values()
                               if hasattr(v, '_rfm_final')}
 
         # Add the final functions from its parents
-        cls._final_methods.update(*(b._final_methods for b in bases
-                                    if hasattr(b, '_final_methods')))
+        bases_w_final = [b for b in bases if hasattr(b, '_final_methods')]
+        cls._final_methods.update(*(b._final_methods for b in bases_w_final))
 
-        if hasattr(cls, '_rfm_special_test') and cls._rfm_special_test:
+        if getattr(cls, '_rfm_override_final', None):
             return
 
         for v in namespace.values():
-            for b in bases:
-                if not hasattr(b, '_final_methods'):
-                    continue
-
+            for b in bases_w_final:
                 if callable(v) and v.__name__ in b._final_methods:
                     msg = (f"'{cls.__qualname__}.{v.__name__}' attempts to "
                            f"override final method "
@@ -303,60 +322,96 @@ class RegressionTestMeta(type):
                     raise ReframeSyntaxError(msg)
 
     def __call__(cls, *args, **kwargs):
-        '''Intercept reframe-specific constructor arguments.
+        '''Inject parameter and variable spaces during object construction.
 
-        When registering a regression test using any supported decorator,
-        this decorator may pass additional arguments to the class constructor
-        to perform specific reframe-internal actions. This gives extra control
-        over the class instantiation process, allowing reframe to instantiate
-        the regression test class differently if this class was registered or
-        not (e.g. when deep-copying a regression test object). These interal
-        arguments must be intercepted before the object initialization, since
-        these would otherwise affect the __init__ method's signature, and these
-        internal mechanisms must be fully transparent to the user.
+        When a class is instantiated, this method intercepts the arguments
+        associated to the parameter and variable spaces. This prevents both
+        :func:`__new__` and :func:`__init__` methods from ever seing these
+        arguments.
+
+        The parameter and variable spaces are injected into the object after
+        construction and before initialization.
         '''
 
+        # Intercept constructor arguments
+        _rfm_use_params = kwargs.pop('_rfm_use_params', False)
+
         obj = cls.__new__(cls, *args, **kwargs)
 
-        # Intercept constructor arguments
-        kwargs.pop('_rfm_use_params', None)
+        # Insert the var & param spaces
+        cls._rfm_var_space.inject(obj, cls)
+        cls._rfm_param_space.inject(obj, cls, _rfm_use_params)
 
         obj.__init__(*args, **kwargs)
         return obj
 
-    def __getattr__(cls, name):
-        '''Attribute lookup method for the MetaNamespace.
+    def __getattribute__(cls, name):
+        '''Attribute lookup method for custom class attributes.
 
-        This metaclass uses a custom namespace, where ``variable`` built-in
-        and ``parameter`` types are stored in their own sub-namespaces (see
-        :class:`reframe.core.meta.RegressionTestMeta.MetaNamespace`). This
-        method will perform an attribute lookup on these sub-namespaces if a
-        call to the default :func:`__getattribute__` method fails to retrieve
-        the requested class attribute.
+        ReFrame test variables are descriptors injected at the class level.
+        If a variable descriptor has already been injected into the class,
+        do not return the descriptor object and return the default value
+        associated with that variable instead.
 
+        .. warning::
+            .. versionchanged:: 3.7.0
+               Prior versions exposed the variable descriptor object if this
+               was already present in the class, instead of returning the
+               variable's default value.
         '''
 
         try:
-            return cls._rfm_var_space.vars[name]
+            var_space = super().__getattribute__('_rfm_var_space')
+        except AttributeError:
+            var_space = None
+
+        # If the variable is already injected, delegate lookup to __getattr__.
+        if var_space and name in var_space.injected_vars:
+            raise AttributeError('delegate variable lookup to __getattr__')
+
+        # Default back to the base method if no special treatment required.
+        return super().__getattribute__(name)
+
+    def __getattr__(cls, name):
+        '''Backup attribute lookup method into custom namespaces.
+
+        Some ReFrame built-in types are stored under their own sub-namespaces.
+        This method will perform an attribute lookup on these sub-namespaces
+        if a call to the default :func:`__getattribute__` method fails to
+        retrieve the requested class attribute.
+        '''
+
+        try:
+            var_space = super().__getattribute__('_rfm_var_space')
+            return var_space.vars[name]
+        except AttributeError:
+            '''Catch early access attempt to the variable space.'''
         except KeyError:
-            try:
-                return cls._rfm_param_space.params[name]
-            except KeyError:
-                raise AttributeError(
-                    f'class {cls.__qualname__!r} has no attribute {name!r}'
-                ) from None
+            '''Requested name not in variable space.'''
+
+        try:
+            param_space = super().__getattribute__('_rfm_param_space')
+            return param_space.params[name]
+        except AttributeError:
+            '''Catch early access attempt to the parameter space.'''
+        except KeyError:
+            '''Requested name not in parameter space.'''
+
+        raise AttributeError(
+            f'class {cls.__qualname__!r} has no attribute {name!r}'
+        ) from None
 
     def __setattr__(cls, name, value):
         '''Handle the special treatment required for variables and parameters.
 
         A variable's default value can be updated when accessed as a regular
-        class attribute. This behaviour does not apply when the assigned value
+        class attribute. This behavior does not apply when the assigned value
         is a descriptor object. In that case, the task of setting the value is
         delegated to the base :func:`__setattr__` (this is to comply with
-        standard Python behaviour). However, since the variables are already
+        standard Python behavior). However, since the variables are already
         descriptors which are injected during class instantiation, we disallow
         any attempt to override this descriptor (since it would be silently
-        re-overriden in any case).
+        re-overridden in any case).
 
         Altering the value of a parameter when accessed as a class attribute
         is not allowed. This would break the parameter space internals.
@@ -400,7 +455,7 @@ class RegressionTestMeta(type):
         This is the case when some parameters are undefined, which results in
         the length of the parameter space being 0.
 
-        :return: bool indicating wheteher the test has undefined parameters.
+        :return: bool indicating whether the test has undefined parameters.
 
         :meta private:
         '''
