@@ -206,7 +206,7 @@ class TestStats:
         run_report = self.json()[-1]
         last_run = run_report['runid']
         for r in run_report['testcases']:
-            if r['result'] == 'success' or r['result'] == 'aborted':
+            if r['result'] in {'success', 'aborted', 'skipped'}:
                 continue
 
             retry_info = (
