@@ -573,7 +573,8 @@ def find_modules(substr, environ_mapping=None):
        class MyTest(rfm.RegressionTest):
            module_info = parameter(find_modules('netcdf'))
 
-           def __init__(self):
+           @rfm.run_after('init')
+           def apply_module_info(self):
                s, e, m = self.module_info
                self.valid_systems = [s]
                self.valid_prog_environs = [e]
@@ -598,7 +599,8 @@ def find_modules(substr, environ_mapping=None):
        class MyTest(rfm.RegressionTest):
            module_info = parameter(my_find_modules('GROMACS'))
 
-           def __init__(self):
+           @rfm.run_after('init')
+           def apply_module_info(self):
                s, e, m = self.module_info
                self.valid_systems = [s]
                self.valid_prog_environs = [e]
@@ -628,8 +630,8 @@ def find_modules(substr, environ_mapping=None):
         if environ_mapping is None:
             return True
 
-        for patt, environs in environ_mapping.items():
-            if re.match(patt, m) and e in environs:
+        for patt, env in environ_mapping.items():
+            if re.match(patt, m) and e == env:
                 return True
 
         return False
